/*----------------------------------------------------------------
readMeta.c
Written by Peter Shawhan, April 2001
Uses the "metaio" library from Philip Charlton
In Matlab, type 'readMeta' without any arguments to see usage info
Modified 12/30/02 by PSS to parse spectrum blobs according to mimetype.
Modified 04/21/05 by PSS to store INT_8S values in Matlab mxINT64 class.

Ideas for future improvements:
*  Allow field name to be something other than the column name.
      (I can't see how to rename an existing field in Matlab)

----------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mex.h"
#include "metaio.h"

void  mexFunction( int nlhs,       mxArray *plhs[],
                   int nrhs, const mxArray *prhs[])
{
  int status;
  int ncols, icol, nusecols, iusecol;
  int icolSpectrum, icolMimetype, mimetypeCode;
  char *mimetype;
  int collist[METAIOMAXCOLS];
  int nrows, nread, irow;

  int allrows, bailed, allcols;
  double *rowlist;
  int nrowlist, irowlist;

  char colspec[1024];

  int datasize[METAIOMAXCOLS], dataused[METAIOMAXCOLS];
  void *dataptr[METAIOMAXCOLS], *newptr;
  int size, length, i, j;
  char file[256], tablename[64];
  int argadj;
  char msg[256];
  char *cptr, *cptr2, *vptr, *dptr;
  char dsave;
  unsigned char *ucptr;
  unsigned char *tempptr;
  unsigned char usave;
  float *fptr;
  const char *colnames_p[METAIOMAXCOLS];
  int coltype[METAIOMAXCOLS];

  mxArray *fout, *carr;
  mxArray **mxap;
  mxChar *mxcptr;
  double *mxrptr;
  double *mxiptr;
  int dims[2], dims2[2];

  struct MetaioParseEnvironment parseEnv;
  const MetaioParseEnv env = &parseEnv;
  struct MetaioRowElement *elt;

  /*-- Byte-swapping unit --*/
  int swap4, swap8;
  union swapcheck_tag {
    int ival;
    char sval[4];
  } swapcheck;

  /*------ Set default parameter values ------*/
  allrows = 1;
  bailed = 0;
  allcols = 1;

  /*------ Check validity of arguments, and get their values ------*/
  if ( nrhs == 0 || nrhs > 4 ) {
    mexErrMsgTxt( "Usage:  a = readMeta( file [,table_name] [,row [,col]] )"
		  "\n  Type 'help readMeta' for details." );
    return;
  }

  if( ! mxIsChar(prhs[0]) ) {
    mexErrMsgTxt( "The first argument must be a filename"
		  "\n  Type 'help readMeta' for usage information." );
    return;
  }
  status = mxGetString( prhs[0], file, sizeof(file) );
  if( status != 0 ) {
    mexErrMsgTxt( "Error getting filename"
		  "\n  Type 'help readMeta' for usage information." );
    return;
  }

  tablename[0] = '\0';
  argadj = 0;

  if ( nrhs > 1 ) {
    if ( mxIsChar(prhs[1]) ) {
      /*-- The user specified a table name as the (optional) 2nd argument --*/
      status = mxGetString( prhs[1], tablename, sizeof(tablename) );
      if( status != 0 ) {
	mexErrMsgTxt( "Error getting table name"
		      "\n  Type 'help readMeta' for usage information." );
	return;
      }
      argadj = 1;
    }
  }

  if ( nrhs > 1+argadj ) {
    if( mxIsChar(prhs[1+argadj]) ) {
      mexErrMsgTxt( "Found a string where the row number(s) should be"
		    "\n  Type 'help readMeta' for usage information." );
      return;
    }

    allrows = 0;
    rowlist = mxGetPr( prhs[1+argadj] );
    nrowlist = mxGetM( prhs[1+argadj] ) * mxGetN( prhs[1+argadj] );
    irowlist = 0;
    /* If the row list is the scalar 0, then return all rows */
    if ( nrowlist==1 && (int)*rowlist == 0 ) { allrows = 1; }
  }

  if ( nrhs > 2+argadj ) {
    if( ! mxIsChar(prhs[2+argadj]) ) {
      mexErrMsgTxt( "The third argument must be a string containing column name(s)"
		    "\n  Type 'help readMeta' for usage information." );
      return;
    }

    status = mxGetString( prhs[2+argadj], colspec, sizeof(colspec) );
    if( status != 0 ) {
      mexErrMsgTxt( "Error getting column specification"
		    "\n  Type 'help readMeta' for usage information." );
      return;
    }
    allcols = 0;
  }

  /*------ open file ------*/
  /*mexPrintf( "Opening file %s\n", file );*/

  status = MetaioOpenTable( env, file, tablename );
  if ( status != 0 ) {
    if ( strlen(tablename) ) {
      sprintf( msg, "Error opening table %s in file %s", tablename, file );
    } else {
      sprintf( msg, "Error opening file %s", file );
    }
    strcat( msg, ": " );
    strcat( msg, env->mierrmsg.data );
    mexErrMsgTxt( msg );
    MetaioAbort( env );
    return;
  }


  /*-- Clear data pointers for all columns --*/
  ncols = env->ligo_lw.table.numcols;
  for ( icol=0; icol<ncols; icol++ ) {
    dataptr[icol] = NULL;
  }


  /*------ Figure out which columns to process ------*/

  if ( allcols == 1 ) {

    nusecols = ncols;
    for ( icol=0; icol < ncols; icol++ ) {
      collist[icol] = icol;
    }

  } else {

    nusecols = 0;

    /*-- Loop over tokens separated by commas and/or spaces.
      Operate on the string in place, rather than copying it. --*/
    vptr = colspec;    /*-- vptr points to the current token --*/
    do {
      /*-- Find the position of the next delimiter (if any) --*/
      dptr = strpbrk( vptr, ", " );
      if ( dptr != NULL ) {
	/*-- If the first character is a delimiter, just continue --*/
	if ( dptr == vptr ) { vptr++; continue; }

	/*-- Temporarily replace the delimiter with a null character,
	  to form a null-terminated token --*/
	dsave = *dptr;
	*dptr = '\0';
      }

      if ( strlen(vptr) > 0 ) {
	/*-- Find the index of the column with the matching name --*/
	icol = MetaioFindColumn( env, vptr );
	if ( icol < 0 ) {
	  sprintf( msg, "Error: column does not exist: %s", vptr );
	  mexErrMsgTxt( msg );
	  MetaioAbort(env);
	  return;
	}

	/*-- Append this column index to the list --*/
	collist[nusecols] = icol;
	nusecols++;
      }

      if ( dptr != NULL ) {
	/*-- Restore the delimiter and get ready to parse next token --*/
	*dptr = dsave;
	vptr = dptr + 1;
      }

    } while ( dptr != NULL );

  }

  /*-- Locate the "spectrum" and "mimetype" columns, if they exist --*/
  icolSpectrum = MetaioFindColumn( env, "spectrum" );
  icolMimetype = MetaioFindColumn( env, "mimetype" );


  /*------ Initialize space to store values read from the table ------*/

  for ( iusecol=0; iusecol<nusecols; iusecol++ ) {
    icol = collist[iusecol];

    /* Get a pointer to the column name, and store it in an array for Matlab */
    cptr = MetaioColumnName( env, icol );
    colnames_p[iusecol] = cptr;

    /* Convert the column name to lowercase in-place! */
    cptr2 = cptr;
    while ( *cptr2 != '\0' ) {
      *cptr2 = (char) tolower( (int) *cptr2 );
      cptr2++;
    }

    coltype[icol] = env->ligo_lw.table.col[icol].data_type;
    if ( nlhs == 0 ) {
      mexPrintf( "  %-20s %s\n", cptr, MetaioTypeText(coltype[icol]) );
    }

    if ( nlhs > 0 ) {
      /*-- Allocate some memory for values from this column --*/
      datasize[icol] = 1024;
      dataused[icol] = 0;
      dataptr[icol] = mxMalloc( datasize[icol] );
      if ( dataptr[icol] == NULL ) {
	mexErrMsgTxt( "Error allocating memory" );
	MetaioAbort( env );
	return;
      }
    }
  }


  /*------ Loop over rows in the file ------*/

  nrows = 0;
  nread = 0;
  while ( (status=MetaioGetRow(env)) == 1 ) {
    nrows++;

    /*-- Make sure we should process this row --*/
    if ( allrows == 0 ) {
      /* If we've read the last specified row, exit from the loop */
      if ( irowlist >= nrowlist ) {
	/* Immediately close the input file */
	MetaioAbort( env );
	bailed = 1;
	break;
      }

      if ( nrows != (int)rowlist[irowlist] ) { continue; }
      /* Increment row-list counter */
      irowlist++;
    }

    nread++;

    /* If there is no output variable, skip the rest of the loop */
    if ( nlhs == 0 ) { continue; }

    irow = nread - 1;
    for ( iusecol=0; iusecol<nusecols; iusecol++ ) {
      icol = collist[iusecol];

      /* All numeric values except INT_8S get converted to double */
      switch ( coltype[icol] ) {
      case METAIO_TYPE_INT_4S:
      case METAIO_TYPE_INT_2S:
      case METAIO_TYPE_REAL_4:
      case METAIO_TYPE_REAL_8:
	size = 8;
	break;
      case METAIO_TYPE_INT_8S:
	size = 8;
	break;
      case METAIO_TYPE_ILWD_CHAR:
      case METAIO_TYPE_LSTRING:
      case METAIO_TYPE_ILWD_CHAR_U:
	size = sizeof(mxArray *);
	break;
      default:
	size = 0;
      }

      dataused[icol] += size;
      if ( dataused[icol] > datasize[icol] ) {
	datasize[icol] *= 2;
	/*mexPrintf( "Calling mxRealloc for column %d\n", icol );*/
	newptr = mxRealloc( dataptr[icol], datasize[icol] );
	if ( newptr == NULL ) {
	  mexErrMsgTxt( "Error allocating more memory" );
	  MetaioAbort( env );
	  return;
	}
	dataptr[icol] = newptr;
      }

      elt = &(env->ligo_lw.table.elt[icol]);

      /*------ For numeric variables, store the value.  For character-array
	variables, create a Matlab array to hold the value, and store a
	pointer to the Matlab array. ------*/

      switch ( coltype[icol] ) {
      case METAIO_TYPE_INT_4S:
	((double *)dataptr[icol])[irow] = (double) elt->data.int_4s;
	break;
      case METAIO_TYPE_INT_2S:
	((double *)dataptr[icol])[irow] = (double) elt->data.int_2s;
	break;
      case METAIO_TYPE_INT_8S:
	((METAIO_INT_8S *)dataptr[icol])[irow] = elt->data.int_8s;
	break;
      case METAIO_TYPE_REAL_4:
	((double *)dataptr[icol])[irow] = (double) elt->data.real_4;
	break;
      case METAIO_TYPE_REAL_8:
	((double *)dataptr[icol])[irow] = elt->data.real_8;
	break;

      case METAIO_TYPE_LSTRING:

	cptr = elt->data.lstring.data;
	length = strlen( cptr );

	/* Trim spaces from the end of the string */
	cptr2 = cptr + length - 1;
	while ( length>0 && *cptr2==' ' ) {
	  *cptr2 = '\0';
	  cptr2--;
	  length--;
	}

	((mxArray **)dataptr[icol])[irow] =
	  mxCreateString( elt->data.lstring.data );
	break;

      case METAIO_TYPE_ILWD_CHAR:

	length = elt->data.ilwd_char.len;
	cptr = (char *) elt->data.ilwd_char.data;

	/* Create a Matlab array to hold the string */
        dims2[0] = 1;
	dims2[1] = length;
	carr = mxCreateCharArray( 2, dims2 );

	/* Store a pointer to this Matlab array */
	((mxArray **)dataptr[icol])[irow] = carr;

	/* Insert the string data into the array */
	mxcptr = (mxChar *) mxGetData(carr);
	for ( i=0; i<length; i++ ) {
	  mxcptr[i] = (mxChar) cptr[i];
	}
	break;

      case METAIO_TYPE_ILWD_CHAR_U:
	length = elt->data.ilwd_char_u.len;
	ucptr = elt->data.ilwd_char_u.data;

	/*-- If this is the "spectrum" column, and there is also a "mimetype"
	  column, and we know how to handle the mimetype, then unpack the
	  numeric values in the BLOB.  Otherwise, just store a char array. --*/
	
	mimetypeCode = 0;
	if ( icol == icolSpectrum && icolMimetype >= 0 ) {
	  mimetype = env->ligo_lw.table.elt[icolMimetype].data.lstring.data;
	  /*mexPrintf( "Mimetype is %s\n", mimetype );*/

	  /*-- See if this is a mimetype we know how to handle, and decide
	    whether to swap bytes (in 4's or 8's) --*/
	  swap4 = 0;
	  swap8 = 0;
	  swapcheck.ival = 1;
	  if ( strcmp(mimetype,"x-ligo/real-spectrum-big") == 0 ) {
	    mimetypeCode = 1;
	    swap4 = ( swapcheck.sval[0] == 1 );
	  } else if ( strcmp(mimetype,"x-ligo/real-spectrum-little") == 0 ) {
	    mimetypeCode = 2;
	    swap4 = ( swapcheck.sval[3] == 1 );
	  } else if ( strcmp(mimetype,"x-ligo/complex-spectrum-big") == 0 ) {
	    mimetypeCode = 3;
	    swap4 = ( swapcheck.sval[0] == 1 );
	  } else if ( strcmp(mimetype,"x-ligo/complex-spectrum-little") ==0 ) {
	    mimetypeCode = 4;
	    swap4 = ( swapcheck.sval[3] == 1 );
	  }

	  /*-- If appropriate, swap bytes in-place --*/
	  if ( swap4 ) {
	    /*-- If length not a multiple of 4, don't swap "leftover" bytes -*/
	    for ( i=0, tempptr=ucptr; i<length-3; i+=4, tempptr+=4 ) {
	      usave = tempptr[0];
	      tempptr[0] = tempptr[3];
	      tempptr[3] = usave;
	      usave = tempptr[1];
	      tempptr[1] = tempptr[2];
	      tempptr[2] = usave;
	    }
	  }

	  if ( swap8 ) {
	    /*-- If length not a multiple of 8, don't swap "leftover" bytes -*/
	    for ( i=0, tempptr=ucptr; i<length-7; i+=8, tempptr+=8 ) {
	      usave = tempptr[0];
	      tempptr[0] = tempptr[7];
	      tempptr[7] = usave;
	      usave = tempptr[1];
	      tempptr[1] = tempptr[6];
	      tempptr[6] = usave;
	      usave = tempptr[2];
	      tempptr[2] = tempptr[5];
	      tempptr[5] = usave;
	      usave = tempptr[3];
	      tempptr[3] = tempptr[4];
	      tempptr[4] = usave;
	    }
	  }

	}

	/*-- Unpack known mimetypes --*/
	switch ( mimetypeCode ) {

	case 1:
	case 2:
	  /*-- Interpret the data as an array of floats --*/
	  fptr = (float *) ucptr;
	  length = length / 4;
	  /*-- Create a Matlab array of doubles --*/
	  carr = mxCreateDoubleMatrix( length, 1, mxREAL );
	  mxrptr = mxGetPr(carr);
	  /*-- Fill the Matlab array, converting floats to doubles --*/
	  for ( i=0; i<length; i++ ) {
	    mxrptr[i] = (double) fptr[i];
	  }
	  break;

	case 3:
	case 4:
	  /*-- Interpret the data as an array of single-precision complex
	    numbers, with (real, imaginary) parts stored together --*/
	  fptr = (float *) ucptr;
	  length = length / 8;
	  /*-- Create a Matlab array of double-precision complex value --*/
	  carr = mxCreateDoubleMatrix( length, 1, mxCOMPLEX );
	  /*-- Get the addresses of the arrays of real and imaginary parts --*/
	  mxrptr = mxGetPr(carr);
	  mxiptr = mxGetPi(carr);
	  /*-- Fill the Matlab arrays, converting floats to doubles --*/
	  for ( i=0, j=0; i<length; i++, j+=2 ) {
	    mxrptr[i] = (double) fptr[j];
	    mxiptr[i] = (double) fptr[j+1];
	  }
	  break;

	default:
	  /*-- Store as a char array --*/
	  dims2[0] = 1;
	  dims2[1] = length;
	  carr = mxCreateCharArray( 2, dims2 );
	  mxcptr = (mxChar *) mxGetData(carr);
	  for ( i=0; i<length; i++ ) {
	    mxcptr[i] = (mxChar) ucptr[i];
	  }

	}

	/*-- Store the address of the Matlab array --*/
	((mxArray **)dataptr[icol])[irow] = carr;

	break;
      }

    }    /*-- End of loop over columns --*/
 }    /*-- End of loop over rows in the file --*/

  /*-- Check for an error --*/
  if ( status == -1 ) {
    mexPrintf( "Parsing error at row %d !\n", nrows+1 );
    mexPrintf( "%s\n", env->mierrmsg.data );
    bailed = 1;
  }

  /*-- Print out number of rows --*/
  if ( nread == 1 ) {
    mexPrintf( "Read 1 row" );
  } else {
    mexPrintf( "Read %d rows", nread );
  }
  if ( allrows == 0 && bailed==0 ) {
    mexPrintf( " out of %d in file\n", nrows );
  } else {
    mexPrintf( "\n" );
  }

  if ( nlhs == 0 ) {
    MetaioClose(env);
    return;
  }


  /*------ Create the output structure ------*/

  /* Create a 1x1 struct matrix for output */
  plhs[0] = mxCreateStructMatrix( 1, 1, nusecols, colnames_p );

  /* We can now close the input file since we don't need colnames_p any more */
  MetaioClose(env);

  /* Create an array for each column, and add it to the structure */
  for ( iusecol=0; iusecol<nusecols; iusecol++ ) {
    icol = collist[iusecol];

    dims[0] = nread;
    dims[1] = 1;

    switch ( coltype[icol] ) {
    case METAIO_TYPE_INT_4S:
    case METAIO_TYPE_INT_2S:
    case METAIO_TYPE_REAL_4:
    case METAIO_TYPE_REAL_8:
      fout = mxCreateDoubleMatrix( nread, 1, mxREAL );
      memcpy( mxGetPr(fout), dataptr[icol], 8*nread );
      break;
    case METAIO_TYPE_INT_8S:
      fout = mxCreateNumericMatrix( nread, 1, mxINT64_CLASS, mxREAL );
      memcpy( mxGetData(fout), dataptr[icol], 8*nread );
      break;
    case METAIO_TYPE_LSTRING:
    case METAIO_TYPE_ILWD_CHAR:
    case METAIO_TYPE_ILWD_CHAR_U:
      fout = mxCreateCellArray( 2, dims );
      mxap = (mxArray **) dataptr[icol];
      for ( irow=0; irow<nread; irow++ ) {
	mxSetCell( fout, irow, mxap[irow] );
      }
      break;
    }

    /* Set each field in output structure */
    mxSetFieldByNumber( plhs[0], 0, iusecol, fout );

    /* Free the temporary array, which is no longer needed */
    mxFree( dataptr[icol] );

  }

  return;
}
